@@ -39,6 +39,7 @@
 #include <osmocom/core/timer.h>
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/bits.h>
+#include <osmocom/core/fsm.h>
 
 #include <osmo-bts/phy_link.h>
 #include <osmo-bts/logging.h>
@@ -47,6 +48,7 @@
 
 #include "l1_if.h"
 #include "trx_if.h"
+#include "trx_provision_fsm.h"
 
 /*
  * socket helper functions
@@ -121,8 +123,8 @@ static int trx_clk_read_cb(struct osmo_fd *ofd, unsigned int what)
 
 	LOGPPHI(pinst, DTRX, LOGL_INFO, "Clock indication: fn=%u\n", fn);
 
-	if (fn >= GSM_HYPERFRAME) {
-		fn %= GSM_HYPERFRAME;
+	if (fn >= GSM_TDMA_HYPERFRAME) {
+		fn %= GSM_TDMA_HYPERFRAME;
 		LOGPPHI(pinst, DTRX, LOGL_ERROR, "Indicated clock's FN is not "
 			"wrapping correctly, correcting to fn=%u\n", fn);
 	}
@@ -202,7 +204,7 @@ void trx_if_init(struct trx_l1h *l1h)
  *  \param[in] fmt Format string (+ variable list of arguments)
  *  \returns 0 on success; negative on error
  *
- *  The new ocommand will be added to the end of the control command
+ *  The new command will be added to the end of the control command
  *  queue.
  */
 static int trx_ctrl_cmd_cb(struct trx_l1h *l1h, int critical, void *cb, const char *cmd,
@@ -211,9 +213,6 @@ static int trx_ctrl_cmd_cb(struct trx_l1h *l1h, int critical, void *cb, const ch
 	struct trx_ctrl_msg *tcm;
 	struct trx_ctrl_msg *prev = NULL;
 	va_list ap;
-	int pending;
-
-	pending = !llist_empty(&l1h->trx_ctrl_list);
 
 	/* create message */
 	tcm = talloc_zero(tall_bts_ctx, struct trx_ctrl_msg);
@@ -236,18 +235,22 @@ static int trx_ctrl_cmd_cb(struct trx_l1h *l1h, int critical, void *cb, const ch
 	tcm->cb = cb;
 
 	/* Avoid adding consecutive duplicate messages, eg: two consecutive POWEROFF */
-	if(pending)
+	if (!llist_empty(&l1h->trx_ctrl_list))
 		prev = llist_entry(l1h->trx_ctrl_list.prev, struct trx_ctrl_msg, list);
-
-	if (!pending ||
-	    !(strcmp(tcm->cmd, prev->cmd) == 0 && strcmp(tcm->params, prev->params) == 0)) {
-		LOGPPHI(l1h->phy_inst, DTRX, LOGL_INFO, "Enqueuing TRX control command 'CMD %s%s%s'\n",
-			tcm->cmd, tcm->params_len ? " ":"", tcm->params);
-		llist_add_tail(&tcm->list, &l1h->trx_ctrl_list);
+	if (prev != NULL && !strcmp(tcm->cmd, prev->cmd)
+			 && !strcmp(tcm->params, prev->params)) {
+		LOGPPHI(l1h->phy_inst, DTRX, LOGL_DEBUG,
+			"Not sending duplicate command '%s'\n", tcm->cmd);
+		talloc_free(tcm);
+		return 0;
 	}
 
+	LOGPPHI(l1h->phy_inst, DTRX, LOGL_INFO, "Enqueuing TRX control command 'CMD %s%s%s'\n",
+		tcm->cmd, tcm->params_len ? " " : "", tcm->params);
+	llist_add_tail(&tcm->list, &l1h->trx_ctrl_list);
+
 	/* send message, if we didn't already have pending messages */
-	if (!pending)
+	if (prev == NULL)
 		trx_ctrl_send(l1h);
 
 	return 0;
@@ -267,32 +270,24 @@ int trx_if_cmd_poweron(struct trx_l1h *l1h, trx_if_cmd_poweronoff_cb *cb)
 }
 
 /*! Send "SETFORMAT" command to TRX: change TRXD header format version */
-int trx_if_cmd_setformat(struct trx_l1h *l1h, uint8_t ver)
+int trx_if_cmd_setformat(struct trx_l1h *l1h, uint8_t ver, trx_if_cmd_generic_cb *cb)
 {
 	LOGPPHI(l1h->phy_inst, DTRX, LOGL_INFO,
 		"Requesting TRXD header format version %u\n", ver);
 
-	return trx_ctrl_cmd(l1h, 0, "SETFORMAT", "%u", ver);
+	return trx_ctrl_cmd_cb(l1h, 0, cb, "SETFORMAT", "%u", ver);
 }
 
 /*! Send "SETTSC" command to TRX */
-int trx_if_cmd_settsc(struct trx_l1h *l1h, uint8_t tsc)
+int trx_if_cmd_settsc(struct trx_l1h *l1h, uint8_t tsc, trx_if_cmd_generic_cb *cb)
 {
-	struct phy_instance *pinst = l1h->phy_inst;
-	if (pinst->phy_link->u.osmotrx.use_legacy_setbsic)
-		return 0;
-
-	return trx_ctrl_cmd(l1h, 1, "SETTSC", "%d", tsc);
+	return trx_ctrl_cmd_cb(l1h, 1, cb, "SETTSC", "%d", tsc);
 }
 
 /*! Send "SETBSIC" command to TRX */
-int trx_if_cmd_setbsic(struct trx_l1h *l1h, uint8_t bsic)
+int trx_if_cmd_setbsic(struct trx_l1h *l1h, uint8_t bsic, trx_if_cmd_generic_cb *cb)
 {
-	struct phy_instance *pinst = l1h->phy_inst;
-	if (!pinst->phy_link->u.osmotrx.use_legacy_setbsic)
-		return 0;
-
-	return trx_ctrl_cmd(l1h, 1, "SETBSIC", "%d", bsic);
+	return trx_ctrl_cmd_cb(l1h, 1, cb, "SETBSIC", "%d", bsic);
 }
 
 /*! Send "SETRXGAIN" command to TRX */
@@ -301,10 +296,16 @@ int trx_if_cmd_setrxgain(struct trx_l1h *l1h, int db)
 	return trx_ctrl_cmd(l1h, 0, "SETRXGAIN", "%d", db);
 }
 
-/*! Send "SETPOWER" command to TRX */
-int trx_if_cmd_setpower(struct trx_l1h *l1h, int db)
+/*! Send "NOMTXPOWER" command to TRX */
+int trx_if_cmd_getnompower(struct trx_l1h *l1h, trx_if_cmd_getnompower_cb *cb)
 {
-	return trx_ctrl_cmd(l1h, 0, "SETPOWER", "%d", db);
+	return trx_ctrl_cmd_cb(l1h, 1, cb, "NOMTXPOWER", "");
+}
+
+/*! Send "SETPOWER" command to TRX */
+int trx_if_cmd_setpower_att(struct trx_l1h *l1h, int power_att_db, trx_if_cmd_setpower_att_cb *cb)
+{
+	return trx_ctrl_cmd_cb(l1h, 0, cb, "SETPOWER", "%d", power_att_db);
 }
 
 /*! Send "SETMAXDLY" command to TRX, i.e. maximum delay for RACH bursts */
@@ -326,7 +327,7 @@ int trx_if_cmd_setslot(struct trx_l1h *l1h, uint8_t tn, uint8_t type, trx_if_cmd
 }
 
 /*! Send "RXTUNE" command to TRX: Tune Receiver to given ARFCN */
-int trx_if_cmd_rxtune(struct trx_l1h *l1h, uint16_t arfcn)
+int trx_if_cmd_rxtune(struct trx_l1h *l1h, uint16_t arfcn, trx_if_cmd_generic_cb *cb)
 {
 	struct phy_instance *pinst = l1h->phy_inst;
 	uint16_t freq10;
@@ -341,11 +342,11 @@ int trx_if_cmd_rxtune(struct trx_l1h *l1h, uint16_t arfcn)
 		return -ENOTSUP;
 	}
 
-	return trx_ctrl_cmd(l1h, 1, "RXTUNE", "%d", freq10 * 100);
+	return trx_ctrl_cmd_cb(l1h, 1, cb, "RXTUNE", "%d", freq10 * 100);
 }
 
 /*! Send "TXTUNE" command to TRX: Tune Transmitter to given ARFCN */
-int trx_if_cmd_txtune(struct trx_l1h *l1h, uint16_t arfcn)
+int trx_if_cmd_txtune(struct trx_l1h *l1h, uint16_t arfcn, trx_if_cmd_generic_cb *cb)
 {
 	struct phy_instance *pinst = l1h->phy_inst;
 	uint16_t freq10;
@@ -360,7 +361,7 @@ int trx_if_cmd_txtune(struct trx_l1h *l1h, uint16_t arfcn)
 		return -ENOTSUP;
 	}
 
-	return trx_ctrl_cmd(l1h, 1, "TXTUNE", "%d", freq10 * 100);
+	return trx_ctrl_cmd_cb(l1h, 1, cb, "TXTUNE", "%d", freq10 * 100);
 }
 
 /*! Send "HANDOVER" command to TRX: Enable handover RACH Detection on timeslot/sub-slot */
@@ -375,6 +376,12 @@ int trx_if_cmd_nohandover(struct trx_l1h *l1h, uint8_t tn, uint8_t ss)
 	return trx_ctrl_cmd(l1h, 1, "NOHANDOVER", "%d %d", tn, ss);
 }
 
+/*! Send "RFMUTE" command to TRX: Mute or Unmute RF transmission */
+int trx_if_cmd_rfmute(struct trx_l1h *l1h, bool mute)
+{
+	return trx_ctrl_cmd(l1h, 0, "RFMUTE", mute ? "1" : "0");
+}
+
 struct trx_ctrl_rsp {
 	char cmd[50];
 	char params[100];
@@ -507,12 +514,17 @@ static int trx_ctrl_rx_rsp_setslot(struct trx_l1h *l1h, struct trx_ctrl_rsp *rsp
 static int trx_ctrl_rx_rsp_setformat(struct trx_l1h *l1h,
 				     struct trx_ctrl_rsp *rsp)
 {
+	trx_if_cmd_generic_cb *cb;
+
 	/* Old transceivers reject 'SETFORMAT' with 'RSP ERR 1' */
 	if (strcmp(rsp->cmd, "SETFORMAT") != 0) {
 		LOGPPHI(l1h->phy_inst, DTRX, LOGL_NOTICE,
 			"Transceiver rejected the format negotiation command, "
 			"using legacy TRXD header format version (0)\n");
-		l1h->config.trxd_hdr_ver_use = 0;
+		if (rsp->cb) {
+			cb = (trx_if_cmd_generic_cb*) rsp->cb;
+			cb(l1h, 0);
+		}
 		return 0;
 	}
 
@@ -525,24 +537,49 @@ static int trx_ctrl_rx_rsp_setformat(struct trx_l1h *l1h,
 		return -EINVAL;
 	}
 
-	/* Transceiver may suggest a lower version (than requested) */
-	if (rsp->status == l1h->config.trxd_hdr_ver_req) {
-		l1h->config.trxd_hdr_ver_use = rsp->status;
-		LOGPPHI(l1h->phy_inst, DTRX, LOGL_INFO,
-			"Using TRXD header format version %u\n",
-			l1h->config.trxd_hdr_ver_use);
-	} else {
-		LOGPPHI(l1h->phy_inst, DTRX, LOGL_DEBUG,
-			"Transceiver suggests TRXD header version %u (requested %u)\n",
-			rsp->status, l1h->config.trxd_hdr_ver_req);
-		/* Send another SETFORMAT with suggested version */
-		l1h->config.trxd_hdr_ver_req = rsp->status;
-		trx_if_cmd_setformat(l1h, rsp->status);
+	if (rsp->cb) {
+		cb = (trx_if_cmd_generic_cb*) rsp->cb;
+		cb(l1h, rsp->status);
 	}
 
 	return 0;
 }
 
+static int trx_ctrl_rx_rsp_nomtxpower(struct trx_l1h *l1h, struct trx_ctrl_rsp *rsp)
+{
+	trx_if_cmd_getnompower_cb *cb = (trx_if_cmd_getnompower_cb*) rsp->cb;
+	struct phy_instance *pinst = l1h->phy_inst;
+	int nominal_power;
+
+	if (rsp->status)
+		LOGPPHI(pinst, DTRX, LOGL_ERROR, "transceiver NOMTXPOWER failed "
+			"with status %d. If your transceiver doesn't support this "
+			"command, then please set the nominal transmit power manually "
+			"through VTY cmd 'nominal-tx-power'.\n",
+			rsp->status);
+	if (cb) {
+		sscanf(rsp->params, "%d", &nominal_power);
+		cb(l1h, nominal_power, rsp->status);
+	}
+	return 0;
+}
+
+static int trx_ctrl_rx_rsp_setpower(struct trx_l1h *l1h, struct trx_ctrl_rsp *rsp)
+{
+	trx_if_cmd_setpower_att_cb *cb = (trx_if_cmd_setpower_att_cb*) rsp->cb;
+	struct phy_instance *pinst = l1h->phy_inst;
+	int power_att;
+
+	if (rsp->status)
+		LOGPPHI(pinst, DTRX, LOGL_ERROR, "transceiver SETPOWER failed with status %d\n",
+			rsp->status);
+	if (cb) {
+		sscanf(rsp->params, "%d", &power_att);
+		cb(l1h, power_att, rsp->status);
+	}
+	return 0;
+}
+
 /* -EINVAL: unrecoverable error, exit BTS
  * N > 0: try sending originating command again after N seconds
  * 0: Done with response, get originating command out from send queue
@@ -551,6 +588,8 @@ static int trx_ctrl_rx_rsp(struct trx_l1h *l1h,
 			   struct trx_ctrl_rsp *rsp,
 			   struct trx_ctrl_msg *tcm)
 {
+	trx_if_cmd_generic_cb *cb;
+
 	if (strcmp(rsp->cmd, "POWERON") == 0) {
 		return trx_ctrl_rx_rsp_poweron(l1h, rsp);
 	} else if (strcmp(rsp->cmd, "POWEROFF") == 0) {
@@ -561,6 +600,16 @@ static int trx_ctrl_rx_rsp(struct trx_l1h *l1h,
 	 * so that's why we should use tcm instead of rsp. */
 	} else if (strcmp(tcm->cmd, "SETFORMAT") == 0) {
 		return trx_ctrl_rx_rsp_setformat(l1h, rsp);
+	} else if (strcmp(tcm->cmd, "NOMTXPOWER") == 0) {
+		return trx_ctrl_rx_rsp_nomtxpower(l1h, rsp);
+	} else if (strcmp(tcm->cmd, "SETPOWER") == 0) {
+		return trx_ctrl_rx_rsp_setpower(l1h, rsp);
+	}
+
+	/* Generic callback if available */
+	if (rsp->cb) {
+		cb = (trx_if_cmd_generic_cb*) rsp->cb;
+		cb(l1h, rsp->status);
 	}
 
 	if (rsp->status) {
@@ -691,7 +740,7 @@ static int trx_data_handle_hdr_v0(struct trx_l1h *l1h,
 	bi->rssi = -(int8_t)buf[5];
 	bi->toa256 = (int16_t) osmo_load16be(buf + 6);
 
-	if (bi->fn >= GSM_HYPERFRAME) {
+	if (bi->fn >= GSM_TDMA_HYPERFRAME) {
 		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
 			"Illegal TDMA fn=%u\n", bi->fn);
 		return -EINVAL;
@@ -1051,34 +1100,30 @@ skip_burst:
 		hdr_ver, trx_data_desc_msg(&bi));
 
 	/* feed received burst into scheduler code */
-	trx_sched_ul_burst(&l1h->l1s, &bi);
+	trx_sched_route_burst_ind(&bi, &l1h->l1s);
 
 	return 0;
 }
 
 /*! Send burst data for given FN/timeslot to TRX
  *  \param[inout] l1h TRX Layer1 handle referring to TX
- *  \param[in] tn Timeslot Number (0..7)
- *  \param[in] fn GSM Frame Number
- *  \param[in] pwr Transmit Power to use
- *  \param[in] bits Unpacked bits to be transmitted
- *  \param[in] nbits Number of \a bits
+ *  \param[in] br Downlink burst request structure
  *  \returns 0 on success; negative on error */
-int trx_if_send_burst(struct trx_l1h *l1h, uint8_t tn, uint32_t fn, uint8_t pwr,
-	const ubit_t *bits, uint16_t nbits)
+int trx_if_send_burst(struct trx_l1h *l1h, const struct trx_dl_burst_req *br)
 {
 	ssize_t snd_len;
 	uint8_t hdr_ver = l1h->config.trxd_hdr_ver_use;
 	uint8_t buf[TRX_DATA_MSG_MAX_LEN];
 
-	if ((nbits != GSM_BURST_LEN) && (nbits != EGPRS_BURST_LEN)) {
-		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR, "Tx burst length %u invalid\n", nbits);
+	if ((br->burst_len != GSM_BURST_LEN) && (br->burst_len != EGPRS_BURST_LEN)) {
+		LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR, "Tx burst length %zu invalid\n",
+			br->burst_len);
 		return -1;
 	}
 
 	LOGPPHI(l1h->phy_inst, DTRX, LOGL_DEBUG,
-		"Tx burst (hdr_ver=%u): tn=%u fn=%u pwr=%u\n",
-		hdr_ver, tn, fn, pwr);
+		"Tx burst (hdr_ver=%u): tn=%u fn=%u att=%u\n",
+		hdr_ver, br->tn, br->fn, br->att);
 
 	switch (hdr_ver) {
 	case 0:
@@ -1092,19 +1137,16 @@ int trx_if_send_burst(struct trx_l1h *l1h, uint8_t tn, uint32_t fn, uint8_t pwr,
 		return -ENOTSUP;
 	}
 
-	buf[0] = ((hdr_ver & 0x0f) << 4) | tn;
-	buf[1] = (fn >> 24) & 0xff;
-	buf[2] = (fn >> 16) & 0xff;
-	buf[3] = (fn >>  8) & 0xff;
-	buf[4] = (fn >>  0) & 0xff;
-	buf[5] = pwr;
+	buf[0] = ((hdr_ver & 0x0f) << 4) | br->tn;
+	osmo_store32be(br->fn, buf + 1);
+	buf[5] = br->att;
 
 	/* copy ubits {0,1} */
-	memcpy(buf + 6, bits, nbits);
+	memcpy(buf + 6, br->burst, br->burst_len);
 
 	/* we must be sure that TRX is on */
 	if (trx_if_powered(l1h)) {
-		snd_len = send(l1h->trx_ofd_data.fd, buf, nbits + 6, 0);
+		snd_len = send(l1h->trx_ofd_data.fd, buf, br->burst_len + 6, 0);
 		if (snd_len <= 0) {
 			strerror_r(errno, (char *)buf, sizeof(buf));
 			LOGPPHI(l1h->phy_inst, DTRX, LOGL_ERROR,
@@ -1192,12 +1234,6 @@ static int trx_if_open(struct trx_l1h *l1h)
 	if (rc < 0)
 		goto err;
 
-	/* enable all slots */
-	l1h->config.slotmask = 0xff;
-
-	if (pinst->num == 0)
-		trx_if_cmd_poweroff(l1h, NULL);
-
 	return 0;
 
 err:
@@ -1262,8 +1298,10 @@ int bts_model_phy_link_open(struct phy_link *plink)
 
 	/* open the individual instances with their ctrl+data sockets */
 	llist_for_each_entry(pinst, &plink->instances, list) {
+		struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
 		if (trx_phy_inst_open(pinst) < 0)
 			goto cleanup;
+		osmo_fsm_inst_dispatch(l1h->provision_fi, TRX_PROV_EV_OPEN, NULL);
 	}
 
 	return 0;
